#include <pthread.h>
#include <sys/time.h>
#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <stdarg.h>
#include <string.h>
#include <math.h>



/* platform detection: select scheduling scope, and use of pthread_setcontention */
/* define ODINMP_SCOPE to be the same as the contention scope for the threads */
/* define ODINMP_SETCONCURRENCY if we must use pthread_setconcurrency, as on IRIX 6.5 */

#ifdef __sun__ /* SUN Solaris */
  #define ODINMP_SCOPE PTHREAD_SCOPE_SYSTEM
  #undef ODINMP_SETCONCURRENCY
#else
  #ifdef __sgi /* SGI IRIX */
    #define ODINMP_SCOPE PTHREAD_SCOPE_PROCESS
    #define ODINMP_SETCONCURRENCY
  #else /* Default values */
    #define ODINMP_SCOPE PTHREAD_SCOPE_SYSTEM
    #undef ODINMP_SETCONCURRENCY
  #endif /* __sgi__ */
#endif /* __sun__ */


/*MARK*/

odinmp_thread_data_t **odinmp_thread_datas;
odinmp_shared_data_t odinmp_shared_data;
odinmp_team_t odinmp_main_team;

pthread_key_t odinmp_pthread_key;
pthread_mutex_t odinmp_atomic_lock;
pthread_mutex_t odinmp_sync_memory_lock;

int odinmp_num_threads = 0;
int odinmp_num_threads_allocated;
int odinmp_debug = 0;
int odinmp_timing = 0;

#ifdef ODINMP_TIMING
double odinmp_timer_global_preStartTime;
double odinmp_timer_global_startTime;
double odinmp_timer_global_stopTime;
double odinmp_timer_wait;

odinmp_tacc_t odinmp_wait_finished_tacc = ODINMP_TACC_INITIALIZER;
#endif /* ODINMP_TIMING */

odinmp_thread_info_t odinmp_thread_init_info;
pthread_mutex_t odinmp_thread_init_lock = PTHREAD_MUTEX_INITIALIZER;
int odinmp_thread_init_working = 0;
pthread_cond_t odinmp_thread_init_done = PTHREAD_COND_INITIALIZER;

odinmp_scheduler_t odinmp_runtime_scheduler = 
{ 
  odinmp_scheduler_static,
  0.0
};

void odinmp_printf(char *fmt, ...) {
    va_list ap;
    va_start(ap, fmt);
    vfprintf(stderr, fmt, ap);
    va_end(ap);
}

odinmp_thread_data_t *odinmp_allocate_thread_data() {
   odinmp_thread_data_t *td = (odinmp_thread_data_t *)
     malloc(sizeof(odinmp_thread_data_t));
   memset(td, 0, sizeof(odinmp_thread_data_t));
   pthread_mutex_init(&ODINMP_THREAD_INFO(td)->lock, NULL);
   pthread_cond_init(&ODINMP_THREAD_INFO(td)->work_available, NULL);
   pthread_cond_init(&ODINMP_THREAD_INFO(td)->finished_work, NULL);
   pthread_cond_init(&ODINMP_THREAD_INFO(td)->stopped_spinning, NULL);   
   ODINMP_THREAD_INFO(td)->team = NULL;
   ODINMP_THREAD_INFO(td)->teams = NULL;
   ODINMP_THREAD_INFO(td)->wsc_info = NULL;
   ODINMP_THREAD_INFO(td)->pthread_id = (pthread_t) 0;
   return td;
}

odinmp_thread_data_t *odinmp_self_data() {
   return (odinmp_thread_data_t *)pthread_getspecific(odinmp_pthread_key);
}

void odinmp_start_thread() {
  pthread_t pthread;
  pthread_attr_t attrs;
  pthread_attr_init(&attrs);
  pthread_attr_setscope(&attrs, ODINMP_SCOPE);
  pthread_create(&pthread, &attrs,
		 odinmp_thread_pthread, NULL);
}

void odinmp_stop_thread(odinmp_thread_data_t *td) {
   ODINMP_TIMING_WAIT_BEGIN;
   pthread_mutex_lock(&ODINMP_THREAD_INFO(td)->lock);
   ODINMP_TIMING_WAIT_END;

   ODINMP_THREAD_INFO(td)->state = odinmp_thread_state_exiting;
   pthread_cond_signal(&ODINMP_THREAD_INFO(td)->work_available);
   pthread_cond_wait(&ODINMP_THREAD_INFO(td)->stopped_spinning,
                     &ODINMP_THREAD_INFO(td)->lock);
   pthread_mutex_unlock(&ODINMP_THREAD_INFO(td)->lock);
}
 
int odinmp_get_num_threads(void) {
  odinmp_thread_info_t *ti = ODINMP_THREAD_INFO(odinmp_self_data());
  return (ti == NULL) ? 1 : ( (ti->team == NULL) ? 1 : ti->team->n );
}

void odinmp_set_num_threads(int num_threads) {
  if (num_threads > odinmp_num_threads_allocated) {
    int i;

    PRINTF(("setting number of threads from %d to %d\n",
	    odinmp_num_threads_allocated, num_threads));

    odinmp_thread_datas = (odinmp_thread_data_t **)
      realloc(odinmp_thread_datas, 
	      num_threads * sizeof(odinmp_thread_data_t *));

    odinmp_main_team.ids = 
      (int *)realloc(odinmp_main_team.ids,
		     num_threads * sizeof(int));

    odinmp_main_team.thread_lap_infos = 
      (odinmp_thread_lap_info_t *)realloc(odinmp_main_team.thread_lap_infos,
					  num_threads * sizeof(odinmp_thread_lap_info_t));

    for (i = odinmp_num_threads_allocated;
	 i < num_threads;
	 i++) {
      pthread_mutex_lock(&odinmp_thread_init_lock);
      odinmp_thread_init_working = 1;
      odinmp_thread_init_info.id = i;
      odinmp_thread_init_info.team_thread_id.id = i;
      odinmp_thread_init_info.team_thread_id.next = NULL;

      odinmp_main_team.ids[i] = i;
      memset(&odinmp_main_team.thread_lap_infos[i], 0, sizeof(odinmp_thread_lap_info_t));

      PRINTF(("thread number %d: id = %d\n",
	       i, odinmp_main_team.ids[i]));

      odinmp_start_thread();
      while (odinmp_thread_init_working) {
	  pthread_cond_wait(&odinmp_thread_init_done, &odinmp_thread_init_lock);
      }
      pthread_mutex_unlock(&odinmp_thread_init_lock);
    }
    odinmp_num_threads_allocated = num_threads;
#ifdef ODINMP_SETCONCURRENCY
    pthread_setconcurrency(odinmp_num_threads_allocated);
#endif
  }
  odinmp_num_threads = num_threads;
  odinmp_main_team.n = num_threads;
}

int odinmp_get_thread_num() {
   odinmp_thread_data_t *td = odinmp_self_data();
   odinmp_team_t *team = ODINMP_THREAD_INFO(td)->team;
   if (team && team == &odinmp_main_team) 
      return ODINMP_THREAD_INFO(td)->id;
   return 0;
}

int odinmp_is_master() {
   odinmp_thread_data_t *td = odinmp_self_data();
   if (td != NULL) {
       odinmp_team_t *team = ODINMP_THREAD_INFO(td)->team;
       if (team != NULL) {
	   return team->ids[0] == ODINMP_THREAD_INFO(td)->id;
       }
   }
   return 1;
}

odinmp_team_t *odinmp_allocate_team(int max_num_threads) {
  odinmp_team_t *team = NULL;

  /* if we are asked for 0 threads, that means we should return
     as many as possible */
  /* also, if we are asked for more threads than we have available,
     we cut down on that number */
  if (max_num_threads == 0 || max_num_threads > odinmp_num_threads)
    max_num_threads = odinmp_num_threads;
      

  if (ODINMP_THREAD_INFO(odinmp_thread_datas[0])->team == NULL) {
    /* main thread in a serial region, return the team that consists */
    /* of all threads - odinmp_main_team */
    int i;
    /* lock each thread, will be unlocked in odinmp_dispatch(); */
    for (i = 0; i < max_num_threads; i++) {
      pthread_mutex_lock(&ODINMP_THREAD_INFO(odinmp_thread_datas[i])->lock);
    }
    odinmp_main_team.n = max_num_threads;
    team = &odinmp_main_team;
    PRINTF(("allocate_team: allocating main team = %x\n", team));
  } else {
    /* we're already parallel, return a team with just the calling */
    /* thread itself in it */
    odinmp_thread_data_t *td = odinmp_self_data();
    odinmp_team_thread_id_list_t *ttil = (odinmp_team_thread_id_list_t *)malloc(sizeof(odinmp_team_thread_id_list_t));

    pthread_mutex_lock(&ODINMP_THREAD_INFO(td)->lock);

    team = (odinmp_team_t *)malloc(sizeof(odinmp_team_t));
    memset(team, 0, sizeof(odinmp_team_t));    
    team->ids = (int *)malloc(sizeof(int));
    team->ids[0] = ODINMP_THREAD_INFO(td)->id;
    team->n = 1;
    team->thread_lap_infos = (odinmp_thread_lap_info_t *)malloc(sizeof(odinmp_thread_lap_info_t));
    memset(&team->thread_lap_infos[0], 0, sizeof(odinmp_thread_lap_info_t));

    ttil->next = ODINMP_THREAD_INFO(td)->team_thread_id.next;
    ttil->id = ODINMP_THREAD_INFO(td)->team_thread_id.id;;
    ODINMP_THREAD_INFO(td)->team_thread_id.next = ttil;
    ODINMP_THREAD_INFO(td)->team_thread_id.id = 0;
    PRINTF(("allocate_team: allocating new team = %x\n", team));
  }

  return team;
}

int odinmp_in_parallel() {
  return (ODINMP_THREAD_INFO(odinmp_self_data())->team != NULL);
}

void odinmp_dispatch(odinmp_thread_data_t *td) {
   /* upon entry to this function, td->lock is still locked */
   /* from odinmp_allocate_team() */
   /* so pthread_mutex_lock(&ODINMP_THREAD_INFO(td)->lock); is not needed */

   ODINMP_THREAD_INFO(td)->state = odinmp_thread_state_has_work_to_do;
   pthread_cond_signal(&ODINMP_THREAD_INFO(td)->work_available);
   pthread_mutex_unlock(&ODINMP_THREAD_INFO(td)->lock);
}

void odinmp_free_team(odinmp_team_t *team) {
   if (team != &odinmp_main_team) {
      odinmp_thread_data_t *td = odinmp_self_data();
      odinmp_team_thread_id_list_t *ttil = ODINMP_THREAD_INFO(td)->team_thread_id.next;
 
      ODINMP_THREAD_INFO(td)->team_thread_id.id = ttil->id;
      ODINMP_THREAD_INFO(td)->team_thread_id.next = ttil->next;
 
      free(ttil);
      free(team->ids);
      free(team->thread_lap_infos);
      free(team);
   }
}

void odinmp_wait_for_work(odinmp_thread_data_t *td) {
   pthread_mutex_lock(&ODINMP_THREAD_INFO(td)->lock);
   while (ODINMP_THREAD_INFO(td)->state != 
	  odinmp_thread_state_has_work_to_do) {
     pthread_cond_wait(&ODINMP_THREAD_INFO(td)->work_available,
		       &ODINMP_THREAD_INFO(td)->lock);
   }
   ODINMP_THREAD_INFO(td)->state = odinmp_thread_state_working;
   pthread_mutex_unlock(&ODINMP_THREAD_INFO(td)->lock);
}

void odinmp_finished_work(odinmp_thread_data_t *td) {
   pthread_mutex_lock(&ODINMP_THREAD_INFO(td)->lock);
   ODINMP_THREAD_INFO(td)->state = odinmp_thread_state_finished_work;
   pthread_cond_signal(&ODINMP_THREAD_INFO(td)->finished_work);
   pthread_mutex_unlock(&ODINMP_THREAD_INFO(td)->lock);
}

void odinmp_wait_finished(odinmp_thread_data_t *td) {
#ifdef ODINMP_TIMING
   double start_time;
#ifdef ODINMP_TIMING_RUNTIME
   if (odinmp_timing) {
#endif /* ODINMP_TIMING_RUNTIME */
   start_time = odinmp_time();
#ifdef ODINMP_TIMING_RUNTIME
   }
#endif /* ODINMP_TIMING_RUNTIME */
#endif /* ODINMP_TIMING */
   pthread_mutex_lock(&ODINMP_THREAD_INFO(td)->lock);
   while (ODINMP_THREAD_INFO(td)->state !=
	  odinmp_thread_state_finished_work) {
      pthread_cond_wait(&ODINMP_THREAD_INFO(td)->finished_work,
                        &ODINMP_THREAD_INFO(td)->lock);
   }
   ODINMP_THREAD_INFO(td)->state = odinmp_thread_state_waiting_for_work;
   pthread_mutex_unlock(&ODINMP_THREAD_INFO(td)->lock);
#ifdef ODINMP_TIMING
#ifdef ODINMP_TIMING_RUNTIME
   if (odinmp_timing) {
#endif /* ODINMP_TIMING_RUNTIME */
   odinmp_tacc_add(&odinmp_wait_finished_tacc, odinmp_time() - start_time);
#ifdef ODINMP_TIMING_RUNTIME
   }
#endif /* ODINMP_TIMING_RUNTIME */
#endif /* ODINMP_TIMING */
}

void odinmp_thread_spin(odinmp_thread_data_t *td) {
   if (ODINMP_THREAD_INFO(td)->id == 0) {
      PRINTF(("thread_spin: thread 0 is main thread, so we don't spin but just return instead\n"));
      return;
   }

   while (1) {
      odinmp_wait_for_work(td);
      if (ODINMP_THREAD_INFO(td)->state ==
	  odinmp_thread_state_exiting) {

	PRINTF(( "spinning for thread %d, was told to stop spinning\n",
		 ODINMP_THREAD_INFO(td)->id,
		 ODINMP_THREAD_INFO(td)->run_region));

         pthread_mutex_lock(&ODINMP_THREAD_INFO(td)->lock);
         pthread_cond_signal(&ODINMP_THREAD_INFO(td)->stopped_spinning);
         pthread_mutex_unlock(&ODINMP_THREAD_INFO(td)->lock);
         return;
      }
      
      PRINTF(( "spinning for thread %d, got work, region %d\n",
	       ODINMP_THREAD_INFO(td)->id,
	       ODINMP_THREAD_INFO(td)->run_region));

      odinmp_thread(td);
      odinmp_finished_work(td);
   }
}

void *odinmp_thread_pthread(void *data) {
   odinmp_thread_data_t *td;

   pthread_mutex_lock(&odinmp_thread_init_lock);
   td = odinmp_thread_datas[odinmp_thread_init_info.id] = odinmp_allocate_thread_data();

   ODINMP_THREAD_INFO(td)->id = odinmp_thread_init_info.id;
   ODINMP_THREAD_INFO(td)->team_thread_id.id = odinmp_thread_init_info.team_thread_id.id;
   ODINMP_THREAD_INFO(td)->team_thread_id.next = odinmp_thread_init_info.team_thread_id.next;

   ODINMP_THREAD_INFO(td)->pthread_id = pthread_self();
   pthread_setspecific(odinmp_pthread_key, td);

   odinmp_thread_init_working = 0;
   pthread_cond_signal(&odinmp_thread_init_done);
   pthread_mutex_unlock(&odinmp_thread_init_lock);

   odinmp_thread_init(td, 0);

   return NULL;
}

void odinmp_init () {
   odinmp_thread_data_t *td;
   int n_threads = ODINMP_NUM_THREADS, i;

   if (odinmp_num_threads > 0) {
     n_threads = odinmp_num_threads;
     odinmp_num_threads = 0;
   }

   pthread_key_create(&odinmp_pthread_key, NULL);

   odinmp_thread_datas = (odinmp_thread_data_t **)
     malloc(sizeof(odinmp_thread_data_t *));
   odinmp_thread_datas[0] = odinmp_allocate_thread_data();
   ODINMP_THREAD_INFO(odinmp_thread_datas[0])->pthread_id = pthread_self();
   odinmp_num_threads_allocated = 1;

   odinmp_main_team.ids = (int *)malloc(sizeof(int));
   odinmp_main_team.thread_lap_infos = 
      (odinmp_thread_lap_info_t *)malloc(sizeof(odinmp_thread_lap_info_t));

   ODINMP_THREAD_INFO(odinmp_thread_datas[0])->id = 0;
   ODINMP_THREAD_INFO(odinmp_thread_datas[0])->team_thread_id.id = 0;
   ODINMP_THREAD_INFO(odinmp_thread_datas[0])->team_thread_id.next = NULL;
   odinmp_main_team.ids[0] = 0;
   memset(&odinmp_main_team.thread_lap_infos[0], 0, sizeof(odinmp_thread_lap_info_t));

   PRINTF(("odinmp preparing %d threads\n", n_threads));

   odinmp_set_num_threads(n_threads);

   pthread_mutex_init(&odinmp_main_team.lock, NULL);

   td = odinmp_thread_datas[0];
   pthread_setspecific(odinmp_pthread_key, td);
   ODINMP_THREAD_INFO(td)->pthread_id = pthread_self();

   /* initialize thread */
   odinmp_thread_init(odinmp_thread_datas[0], 0);

   /* initialize atomic mutex */
   pthread_mutex_init(&odinmp_atomic_lock, NULL);

   /* initialize memory sync mutex */
   pthread_mutex_init(&odinmp_sync_memory_lock, NULL);

   /* initialize generated variables such as mutexes for critical regions */
/*MARK*/
}

void odinmp_atexit () {
/*MARK*/
    return;
}


void odinmp_complain() {
  fprintf(stderr,"Complaint from OdinMP!\n");
}

int odinmp_get_num_procs() {
  return ODINMP_NUM_PROCESSORS;
}

void odinmp_barrier(odinmp_team_t *team, odinmp_barrier_t *b)
{
  /* odinmp_sync_memory(); */
  /* not necessary */
  ODINMP_TIMING_WAIT_BEGIN;
  pthread_mutex_lock(&team->lock);
  ODINMP_TIMING_WAIT_END;

  while (b->state == odinmp_barrier_state_waiting_for_depart) {
    PRINTF(( "thread %d: waiting for empty on barrier %0x, %d threads need to depart\n",
	     ODINMP_THREAD_INFO(odinmp_self_data())->id, 
	     (int)b, b->n_left_to_depart ));
    pthread_cond_wait(&b->all_departed, &team->lock);
  }

  if (b->state == odinmp_barrier_state_empty) {
    
    b->n_left_to_join = b->n_left_to_depart = team->n;
    pthread_cond_init(&b->all_joined, NULL);
    pthread_cond_init(&b->all_departed, NULL);
    b->state = odinmp_barrier_state_waiting_for_join;

    PRINTF(( "thread %d: initializing barrier %0x, waiting for %d threads to join\n",
	     ODINMP_THREAD_INFO(odinmp_self_data())->id, 
	     (int)b,
	     b->n_left_to_join));
    
  }
  
  b->n_left_to_join--;
  PRINTF(( "thread %d: joining barrier %0x, waiting for %d more threads to join\n",
	     ODINMP_THREAD_INFO(odinmp_self_data())->id, 
	     (int)b,
	     b->n_left_to_join));

  if (b->n_left_to_join == 0) {
    PRINTF(( "thread %d: barrier %0x join complete, telling the other threads\n",
	     ODINMP_THREAD_INFO(odinmp_self_data())->id, 
	     (int)b));
    b->state = odinmp_barrier_state_waiting_for_depart;
    pthread_cond_broadcast(&b->all_joined);
  } else {
    PRINTF(( "thread %d: waiting on barrier %0x\n",
	     ODINMP_THREAD_INFO(odinmp_self_data())->id, 
	     (int)b));
    pthread_cond_wait(&b->all_joined, &team->lock);
    PRINTF(( "thread %d: finished waiting on barrier %0x\n",
	     ODINMP_THREAD_INFO(odinmp_self_data())->id, 
	     (int)b));
  }
  if (--b->n_left_to_depart == 0) {
    PRINTF(( "thread %d: last to depart from barrier %0x, telling the other threads\n",
	     ODINMP_THREAD_INFO(odinmp_self_data())->id, 
	     (int)b));
    b->state = odinmp_barrier_state_empty;
    pthread_cond_broadcast(&b->all_departed);
  }

  pthread_mutex_unlock(&team->lock);
}

void odinmp_sync_memory() {
  /* machine independant solution: lock and unlock a mutex */
  ODINMP_TIMING_WAIT_BEGIN;
  pthread_mutex_lock(&odinmp_sync_memory_lock);
  pthread_mutex_unlock(&odinmp_sync_memory_lock);
  ODINMP_TIMING_WAIT_END;
}

double odinmp_time() {
    struct timeval tv;
    gettimeofday(&tv, ((void *)0));
    return (double)tv.tv_sec + (double)tv.tv_usec/1000000.0;
}

void odinmp_tacc_init(odinmp_tacc_t *tacc) {
    pthread_mutex_init(&tacc->mutex, ((void *)0));
    tacc->min = HUGE_VAL;
    tacc->max = -HUGE_VAL;
    tacc->total = 0.0;
    tacc->n = 0;
}

void odinmp_tacc_add(odinmp_tacc_t *tacc, double d) {
    pthread_mutex_lock(&tacc->mutex);
    if (d < tacc->min)
	tacc->min = d;
    if (d > tacc->max)
	tacc->max = d;
    tacc->total += d;
    tacc->n++;
    pthread_mutex_unlock(&tacc->mutex);
}

void odinmp_parse_env () {
    char *val;

    /* handle OMP_SCHEDULE */
    val = getenv("OMP_SCHEDULE");
    if (val != NULL) {
	int chunk_size = 0;
	char *s = strchr(val, ',');
	if (s != NULL) {
	    /* we have a comma, ie "scheduler,chunk_size" */
	    *s++ = '\0';
	    chunk_size = atoi(s);
	}

	if (!strncasecmp(val, "guided", 6)) {
	    /* GUIDED */
	    odinmp_runtime_scheduler.type = odinmp_scheduler_guided;
	    odinmp_runtime_scheduler.chunk_size = chunk_size;
	} else if (!strncasecmp(val, "dynamic", 7)) {
	    /* DYNAMIC */
	    odinmp_runtime_scheduler.type = odinmp_scheduler_dynamic;
	    odinmp_runtime_scheduler.chunk_size = chunk_size;
	} else if (!strncasecmp(val, "static", 6)) {
	    /* STATIC */
	    odinmp_runtime_scheduler.type = odinmp_scheduler_static;
	    odinmp_runtime_scheduler.chunk_size = chunk_size;
	} else {
	    fprintf(stderr,"Unknown OMP_SCHEDULE scheduler \"%s\", ignoring\n",val);
	}
    }

    /* handle OMP_NUM_THREADS */
    val = getenv("OMP_NUM_THREADS");
    if (val != NULL) {
	odinmp_num_threads = atoi(val);
	if (odinmp_num_threads < 0) {
	  odinmp_num_threads = 0; 
	}
    }

    /* handle OMP_DYNAMIC */
    val = getenv("OMP_DYNAMIC");
    if (val != NULL) {
	/* we still don't do anything, we don't implement dynamic */
    }

    /* handle OMP_NESTED */
    val = getenv("OMP_NESTED");
    if (val != NULL) {
	/* we still don't do anything, we don't implement nesting */	
    }

    /* handle ODINMP_DEBUG */
    val = getenv("ODINMP_DEBUG");
    if (val != NULL) {
	odinmp_debug = 1;
    }
}

void odinmp_parse_args (int *argc, char **argv) {
  int argn;
  int nargs = *argc;

  PRINTF(("we have %d arguments:\n", nargs));
  for (argn = 0; argn < *argc; argn++)
    PRINTF(( "``%s''\n", argv[argn]));
  
  for (argn = 0; argn < nargs;) {
    PRINTF(("checking argument %d: ``%s''\n", argn, argv[argn]));
    if (!strncmp(argv[argn], "--odinmp_", 9)) {
      int i;
      char *s = argv[argn];

      /* move all the other arguments together */
      for (i = argn+1; i < nargs; i++)
	argv[i-1] = argv[i];
      argv[nargs--] = ((char *)0);

      /* now, parse this odinmp argument */
      /* first, move ahead 9 steps, skipping the initial "--odinmp_" */
      s += 9;
      if (!strncmp(s, "num_threads=", 12)) {
	/* skip ahead again */
	s += 12;
	odinmp_num_threads = atoi(s);
	if (odinmp_num_threads < 0)
	  odinmp_num_threads = 0;

	PRINTF(("running with %d threads\n", odinmp_num_threads));
      } else if (!strncmp(s, "debug", 5)) {
	odinmp_debug = 1;
	PRINTF(("running with debugging printouts\n"));
      } else if (!strncmp(s, "timing", 6)) {
	odinmp_timing = 1;
	PRINTF(("running with %d levels of timing printouts\n", odinmp_timing));
      } else {
	fprintf(stderr, "unknown odinmp option ``%s''\n", s);
      }
    } else {
      argn++;
    }
  }
  *argc = nargs;
}
