typedef struct _odinmp_team_t odinmp_team_t;
typedef struct _odinmp_thread_data_t odinmp_thread_data_t;
typedef struct _odinmp_shared_data_t odinmp_shared_data_t;

typedef struct {
    double min, max, total;
    int n;
    pthread_mutex_t mutex;
} odinmp_tacc_t;

#define ODINMP_TACC_INITIALIZER { 0.0, 0.0, 0.0, 0, PTHREAD_MUTEX_INITIALIZER }
#define ODINMP_THREAD_INFO(x) ((odinmp_thread_info_t *)(x))

typedef struct {
  long long int lb;
  long long int incr; 
  long long int b; 
  int last, done; 
  int i;
  int in_order;
} odinmp_for_slice_t; 

typedef enum {
  odinmp_scheduler_runtime = 0,
  odinmp_scheduler_static = 1,
  odinmp_scheduler_dynamic = 2,
  odinmp_scheduler_guided = 3
} odinmp_scheduler_type_t;

typedef struct {
  odinmp_scheduler_type_t type;
  double chunk_size;
} odinmp_scheduler_t;

typedef enum {
  odinmp_barrier_state_empty = 0,
  odinmp_barrier_state_waiting_for_join,
  odinmp_barrier_state_waiting_for_depart
} odinmp_barrier_state_t;

typedef struct {
  pthread_cond_t all_joined;
  pthread_cond_t all_departed;
  int n_left_to_join;
  int n_left_to_depart;
  odinmp_barrier_state_t state;
} odinmp_barrier_t;

typedef struct {
  long long int lb;
  long long int incr;
  long long int b;
  double chunk_size;
  odinmp_scheduler_t scheduler;

  pthread_mutex_t order_lock;
  pthread_cond_t new_order;
  int in_order;
  int next_in_order;

} odinmp_for_data_t;

typedef struct _odinmp_for_lap_list_t {
  odinmp_for_data_t for_data;
  int lap;
  int n_left_to_leave;

  struct _odinmp_for_lap_list_t *next;
} odinmp_for_lap_list_t;

typedef struct {
  int section;
  int last;
  int done;
} odinmp_sect_slice_t;

typedef struct {
  int n_left; 
} odinmp_sect_data_t;

typedef struct _odinmp_sect_lap_list_t {
  odinmp_sect_data_t sect_data;
  int lap;
  int n_left_to_leave;

  struct _odinmp_sect_lap_list_t *next;
} odinmp_sect_lap_list_t;

typedef struct {
  int do_it;
} odinmp_sing_slice_t;

typedef struct _odinmp_team_list_t {
  odinmp_team_t *team;
  struct _odinmp_team_list_t *next;
} odinmp_team_list_t; 

enum {
  odinmp_wsc_type_none = 0,
  odinmp_wsc_type_for,
  odinmp_wsc_type_sect,
  odinmp_wsc_type_sing
};

typedef struct {
  int lap;
  int in_order;
} odinmp_for_lap_info_t;

typedef struct {
  int lap;
} odinmp_sect_lap_info_t;

typedef struct {
  int lap;
} odinmp_sing_lap_info_t;

typedef struct _odinmp_wsc_info_t {
  int type;
  struct _odinmp_wsc_info_t *next;

  union {
    struct { 
      odinmp_for_lap_list_t *lap_list;
      odinmp_for_slice_t *slice;
    } wsc_for;

    struct {
      odinmp_sect_lap_list_t *lap_list;
    } wsc_sect;

  } u;

} odinmp_wsc_info_t;

typedef struct _odinmp_team_thread_id_list_t {
  int id;
  struct _odinmp_team_thread_id_list_t *next;
} odinmp_team_thread_id_list_t;

typedef enum {
  odinmp_thread_state_waiting_for_work,
  odinmp_thread_state_has_work_to_do,
  odinmp_thread_state_working,
  odinmp_thread_state_finished_work,
  odinmp_thread_state_exiting
} odinmp_thread_state_t;

typedef struct {
  pthread_mutex_t lock;
  pthread_cond_t work_available;
  pthread_cond_t finished_work;
  pthread_cond_t stopped_spinning;
  pthread_t pthread_id;
  odinmp_thread_state_t state;

  int id;
  int run_region;

  odinmp_wsc_info_t *wsc_info;
  
  odinmp_team_t *team;
  odinmp_team_list_t *teams;
  odinmp_team_thread_id_list_t team_thread_id;

/*MARK*/
} odinmp_thread_info_t; 



/*MARK*/


/* typedef */ struct _odinmp_team_t {
  pthread_mutex_t lock;
  
  int id;
  int *ids;
  int n;
  odinmp_thread_lap_info_t *thread_lap_infos;
  
/*MARK*/
} /* odinmp_team_t */ ;


