#include "omp.h"
#include <pthread.h>

extern int odinmp_num_threads;
extern int odinmp_get_num_threads(void);
extern void odinmp_set_num_threads(int num_threads);
extern int odinmp_get_thread_num();
extern int odinmp_get_num_procs();
extern int odinmp_in_parallel();


void omp_set_num_threads(int num_threads) {
  odinmp_set_num_threads(num_threads);
}

int omp_get_num_threads(void) {
  return odinmp_get_num_threads();
}

int omp_get_max_threads() {
   return odinmp_num_threads;
}

int omp_get_thread_num () {
  return odinmp_get_thread_num();
}

int omp_get_num_procs() {
   return odinmp_get_num_procs();
}

int omp_in_parallel() {
   return odinmp_in_parallel();
}

void omp_set_dynamic(int dymanic_threads) {
   /* we do not support dynamic thread adjustment yet */
   /* do nothing */
}

int omp_get_dynamic() {
   /* we do not support dynamic thread adjustment yet */
   return 0;
}

void omp_set_nested(int nested) {
   /* we do not support nested parallelism yet */
   /* do nothing */
}

int omp_get_nested() {
   /* we do not support nested parallelism yet */
   return 0;
}

/* lock functions */

void omp_init_lock(omp_lock_t *lock) {
  pthread_mutex_init(lock, NULL);
}

void omp_init_nest_lock(omp_nest_lock_t *lock) {
  pthread_mutex_init(&lock->mutex, NULL);
  pthread_cond_init(&lock->released, NULL);
  lock->locked_by = (pthread_t)0;
  lock->depth = 0;
}


void omp_destroy_lock(omp_lock_t *lock) {
  pthread_mutex_destroy(lock);
}

void omp_destroy_nest_lock(omp_nest_lock_t *lock) {
  pthread_mutex_destroy(&lock->mutex);
  pthread_cond_destroy(&lock->released);
}


void omp_set_lock(omp_lock_t *lock) {
  pthread_mutex_lock(lock);
}

void omp_set_nest_lock(omp_nest_lock_t *lock) {
  pthread_mutex_lock(&lock->mutex);
  if (pthread_equal(lock->locked_by, pthread_self())) {
    lock->depth++;
  } else {
    while (lock->depth > 0) {
      pthread_cond_wait(&lock->released, &lock->mutex);
    }
    lock->depth = 1;
    lock->locked_by = pthread_self();
  }
  pthread_mutex_unlock(&lock->mutex);
}


void omp_unset_lock(omp_lock_t *lock) {
  pthread_mutex_unlock(lock);
}

void omp_unset_nest_lock(omp_nest_lock_t *lock) {
  pthread_mutex_lock(&lock->mutex);
  if ((lock->depth --) == 0) {
    pthread_cond_broadcast(&lock->released);
    lock->locked_by = (pthread_t)0;
  }
  pthread_mutex_unlock(&lock->mutex);
}


int omp_test_lock(omp_lock_t *lock) {
  return (pthread_mutex_trylock(lock) == 0);
}

int omp_test_nest_lock(omp_nest_lock_t *lock) {
  int v = 0;
  pthread_mutex_lock(&lock->mutex);
  if (pthread_equal(lock->locked_by, pthread_self())) {
    v = lock->depth += 1;
  } else if (lock->depth == 0) {
    v = lock->depth = 1;
    lock->locked_by = pthread_self();
  } else {
    v = 0;
  }
  pthread_mutex_unlock(&lock->mutex);
  return v;
}

